# Functions from the following NumPy document
# https://numpy.org/doc/stable/reference/routines.polynomials.html

# "NOQA" to suppress flake8 warning
from cupy.polynomial import polynomial  # NOQA
from cupy.polynomial import polyutils  # NOQA
